
import asyncio
import datetime
import logging
from aioesphomeapi import (
    APIClient,
    EntityState,
    ResolveAPIError,
    APIConnectionError,
    InvalidEncryptionKeyAPIError,
    RequiresEncryptionAPIError,
    ReconnectLogic,
)
from zeroconf import Zeroconf

from estimenergy.collectors import Collector
from estimenergy.models import CollectorData, EnergyData
from estimenergy.metrics import CollectorMetrics


class GlowCollector(Collector):
    def __init__(
        self,
        collector_data: CollectorData
    ):
        self.collector_data = collector_data
        self.logger = logging.getLogger("energy_collector").getChild(self.collector_data.name)

        self.logger.info(f"Creating API Client for {self.collector_data.name} ({self.collector_data.host}:{self.collector_data.port})")

        self.zeroconf = Zeroconf()
        self.api = APIClient(
            self.collector_data.host,
            self.collector_data.port,
            self.collector_data.password,
            zeroconf_instance=self.zeroconf,
        )
        self.reconnect_logic = ReconnectLogic(        
            client=self.api,
            on_connect=self.__on_connect,
            on_disconnect=self.__on_disconnect,
            zeroconf_instance=self.zeroconf,
            name=self.collector_data.name,
            on_connect_error=self.__on_connect_error,
        )
        
        self.metrics = CollectorMetrics(self.collector_data)

    async def start(self):
        await self.__try_login()
        await self.reconnect_logic.start()

    async def __try_login(self):
        try:
            await self.api.connect(login=True)
            self.device_info = await self.api.device_info()
        except ResolveAPIError or APIConnectionError or InvalidEncryptionKeyAPIError or RequiresEncryptionAPIError as err:
            raise err
        finally:
            await self.api.disconnect(force=True)

    async def __on_connect(self):
        self.logger.info(f"Connected to ESPHome Device {self.collector_data.name}")

        try:
            await self.api.subscribe_states(self.__state_changed)
        except APIConnectionError as err:
            await self.api.disconnect()
    
    async def __on_disconnect(self):
        self.logger.warn(f"Disconnected from ESPHome Device {self.name}")

    async def __on_connect_error(self, exception: Exception):
        self.logger.error(f"Error connecting to ESPHome Device {self.name}")
        self.logger.error(exception)

    def __state_changed(self, state: EntityState):
        if not state.key == 3673186328:
            return
        
        current_kwh: float = state.state
        loop = asyncio.get_event_loop()
        loop.create_task(self.__on_kwh_changed(current_kwh))
    
    async def __on_kwh_changed(self, current_kwh: float):
        date = datetime.datetime.now()

        self.logger.info(f"Current KWh: {current_kwh}")

        energy_data = await EnergyData.filter(collector=self.collector_data, year=date.year, month=date.month, day=date.day).first()
        
        if energy_data is None:
            energy_data = EnergyData(
                collector=self.collector_data,
                year=date.year,
                month=date.month,
                day=date.day,
                kwh=current_kwh,
                hour_created=date.hour,
                hour_updated=date.hour,
                is_completed=False
            )
            await energy_data.save()
            await self.__update_previous_energy_data(date)
            return

        energy_data.kwh = current_kwh
        energy_data.hour_updated = date.hour
        await energy_data.save()
        await self.metrics.update_metrics()

    async def __update_previous_energy_data(self, date):
        date_yesterday = date - datetime.timedelta(days=1)
        previous_energy_data = await EnergyData.filter(
            collector=self.collector_data,
            year=date_yesterday.year,
            month=date_yesterday.month,
            day=date_yesterday.day
        ).first()

        if previous_energy_data is None:
            return
        
        if previous_energy_data.hour_updated < 23:
            return
        
        previous_energy_data.is_completed = True
        await previous_energy_data.save()




from prometheus_client import CollectorRegistry
from prometheus_fastapi_instrumentator import Instrumentator
from dotenv import load_dotenv

from estimenergy.models.settings import Settings


load_dotenv()

settings = Settings()
instrumentator = Instrumentator(
    should_group_status_codes=False,
    should_ignore_untemplated=True,
    should_respect_env_var=True,
    should_instrument_requests_inprogress=True,
    excluded_handlers=[".*admin.*", "/metrics"],
    env_var_name="ENABLE_METRICS",
    inprogress_name="inprogress",
    inprogress_labels=True,
)



import requests


class EstimEnergyClient:
    def __init__(self, host, port):
        self.host = host
        self.port = port

    def get_collector(self, id):
        url = f"http://{self.host}:{self.port}/collector/{id}"
        response = requests.get(url)
        return response.json()
    
    async def async_get_collector(self, id):
        url = f"http://{self.host}:{self.port}/collector/{id}"
        response = await requests.get(url)
        return response.json()
    
    def get_collectors(self):
        url = f"http://{self.host}:{self.port}/collector"
        response = requests.get(url)
        return response.json()
    
    async def async_get_collectors(self):
        url = f"http://{self.host}:{self.port}/collector"
        response = await requests.get(url)
        return response.json()




from enum import Enum
from prometheus_client import Gauge

DEFAULT_HOST = "localhost"
DEFAULT_PORT = 8000

SENSOR_TYPE_JSON = "json"
SENSOR_TYPE_FRIENDLY_NAME = "friendly_name"
SENSOR_TYPE_UNIQUE_ID = "unique_id"

class MetricPeriod(Enum):
    DAY = ("day", "Daily")
    MONTH = ("month", "Monthly")
    YEAR = ("year", "Yearly")
    TOTAL = ("total", "Total")

class MetricType(Enum):
    COST = ("cost", "Cost")
    COST_DIFFERENCE = ("cost_difference", "Cost Difference")
    ENERGY = ("kwh", "Energy")
    ACCURACY = ("accuracy", "Accuracy")

class Metric:
    def __init__(self, metric_type: MetricType, metric_period: MetricPeriod, is_predicted: bool, is_raw: bool):
        self.metric_type = metric_type
        self.metric_period = metric_period
        self.is_predicted = is_predicted
        self.is_raw = is_raw
        
    @property
    def json_key(self) -> str:
        return f"{self.metric_period.value[0]}_{self.metric_type.value[0]}{'_predicted' if self.is_predicted else ''}{'_raw' if self.is_raw else ''}"

    @property
    def friendly_name(self) -> str:
        return f"{self.metric_period.value[1]} {self.metric_type.value[1]} {'(Predicted)' if self.is_predicted else ''} {'(Raw)' if self.is_raw else ''}"
    
    def create_gauge(self) -> Gauge:
        return Gauge(
            f"estimenergy_{self.json_key}",
            f"EstimEnergy {self.friendly_name}",
            ["name", "id"],
        )

METRICS = [
    Metric(MetricType.ENERGY, MetricPeriod.DAY, False, False),
    Metric(MetricType.ENERGY, MetricPeriod.MONTH, False, False),
    Metric(MetricType.ENERGY, MetricPeriod.YEAR, False, False),
    Metric(MetricType.ENERGY, MetricPeriod.TOTAL, False, False),
    Metric(MetricType.ENERGY, MetricPeriod.MONTH, True, False),
    Metric(MetricType.ENERGY, MetricPeriod.YEAR, True, False),
    Metric(MetricType.ENERGY, MetricPeriod.MONTH, True, True),
    Metric(MetricType.ENERGY, MetricPeriod.YEAR, True, True),
    Metric(MetricType.COST, MetricPeriod.DAY, False, False),
    Metric(MetricType.COST, MetricPeriod.MONTH, False, False),
    Metric(MetricType.COST, MetricPeriod.YEAR, False, False),
    Metric(MetricType.COST, MetricPeriod.MONTH, True, False),
    Metric(MetricType.COST, MetricPeriod.YEAR, True, False),
    Metric(MetricType.COST, MetricPeriod.MONTH, True, True),
    Metric(MetricType.COST, MetricPeriod.YEAR, True, True),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.DAY, False, False),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.MONTH, False, False),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.YEAR, False, False),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.MONTH, True, False),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.YEAR, True, False),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.MONTH, True, True),
    Metric(MetricType.COST_DIFFERENCE, MetricPeriod.YEAR, True, True),
    Metric(MetricType.ACCURACY, MetricPeriod.DAY, False, False),
    Metric(MetricType.ACCURACY, MetricPeriod.MONTH, False, False),
    Metric(MetricType.ACCURACY, MetricPeriod.YEAR, False, False),
]

JSON_NAME = "name"
JSON_HOST = "host"
JSON_PORT = "port"
JSON_COST_PER_KWH = "cost_per_kwh"
JSON_BASE_COST_PER_MONTH = "base_cost_per_month"
JSON_PAYMENT_PER_MONTH = "payment_per_month"
JSON_MIN_HOUR = "min_hour"
JSON_MAX_HOUR = "max_hour"
JSON_MAX_INCOMPLETE_DAYS = "max_incomplete_days"
JSON_BILLING_MONTH = "billing_month"
JSON_DATA = "data"

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "uvicorn.logging.ColourizedFormatter",
            "format": "%(asctime)-25s %(name)-30s %(levelprefix)-8s %(message)s",
        },
        "access": {
            "()": "uvicorn.logging.ColourizedFormatter",
            "format": "%(asctime)-25s %(name)-30s %(levelprefix)-8s %(message)s",
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "access": {
            "formatter": "access",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "uvicorn.error": {
            "level": "INFO",
            "handlers": ["default"],
            "propagate": False,
        },
        "uvicorn.access": {
            "level": "INFO",
            "handlers": ["access"],
            "propagate": False,
        },
        "energy_collector": {
            "level": "DEBUG",
            "handlers": ["default"],
            "propagate": False,
        },
    },
}
